from __future__ import annotations

from collections.abc import Callable
from contextlib import ExitStack
from typing import Generic, TypeVar

from .handler import ErrorHookHandler
from .job import Job
from .job_runner import JobRunner
from .job_state_store import JobStateStoreWorkerProtocol
from .queue import QueueProtocol
from .types import ExceptionHook, Lifespan
from .workflow import Workflow
from .workflow_registry import WorkflowRegistry

AppType = TypeVar("AppType", bound="Ergate")
JobType = TypeVar("JobType", bound=Job)


class Ergate(Generic[JobType]):
    def __init__(
        self: AppType,
        queue: QueueProtocol[JobType],
        job_state_store: JobStateStoreWorkerProtocol[JobType],
        lifespan: Lifespan[AppType] | None = None,
        error_hook_handler: ErrorHookHandler[JobType] | None = None,
    ) -> None:
        self.lifespan = lifespan
        self.error_hook_handler: ErrorHookHandler[JobType] = (
            error_hook_handler or ErrorHookHandler()
        )

        self.workflow_registry = WorkflowRegistry()
        self.job_runner: JobRunner[JobType] = JobRunner(
            queue,
            self.workflow_registry,
            job_state_store,
            self.error_hook_handler,
        )

    def exception_hook(
        self, *exceptions: type[Exception]
    ) -> Callable[[ExceptionHook[JobType]], ExceptionHook[JobType]]:
        def decorator(func: ExceptionHook[JobType]) -> ExceptionHook[JobType]:
            for exception in exceptions:
                self.error_hook_handler.register(exception, func)
            return func

        return decorator

    def register_workflow(self, workflow: Workflow) -> None:
        self.workflow_registry.register(workflow)

    def run(self: AppType) -> None:
        with ExitStack() as stack:
            if self.lifespan:
                stack.enter_context(self.lifespan(self))
            self.job_runner.run()



from collections.abc import Callable
from typing import ContextManager, TypeVar

from .annotations import Context, Depends, Input

AppType = TypeVar("AppType")
JobType = TypeVar("JobType")

Lifespan = Callable[[AppType], ContextManager[None]] | None

ExceptionHook = Callable[[JobType, Exception], None]

Annotation = Input | Depends | Context



from __future__ import annotations

from contextlib import ExitStack, contextmanager
from typing import TYPE_CHECKING, Any, Callable, Generator, Generic, TypeVar

from .depends_cache import DependsCache

if TYPE_CHECKING:
    from .inspect import FunctionArgumentInfo

DependencyReturn = TypeVar("DependencyReturn")


class Depends(Generic[DependencyReturn]):
    def __init__(
        self,
        dependency: Callable[..., Generator[DependencyReturn, None, None]],
    ) -> None:
        self.dependency = dependency
        self.argument_info: FunctionArgumentInfo | None = None

    def initialize(self, argument_info: FunctionArgumentInfo) -> None:
        self.argument_info = argument_info

    @contextmanager
    def create(
        self,
        stack: ExitStack,
        depends_cache: DependsCache,
        user_context: Any,
        input_value: Any,
    ) -> Generator[DependencyReturn, None, None]:
        assert self.argument_info is not None, "Depends not initialized"

        if self.dependency in depends_cache:
            yield depends_cache[self.dependency]
            return

        args, kwargs = self.argument_info.build_args(
            stack,
            depends_cache,
            user_context,
            input_value,
        )

        dependency_callable = contextmanager(self.dependency)
        dependency = stack.enter_context(dependency_callable(*args, **kwargs))
        depends_cache.set(self.dependency, dependency)

        yield dependency


class Input:
    pass


class Context:
    pass



from typing import Any


class DependsCache:
    def __init__(self) -> None:
        self._cache: dict[Any, Any] = {}

    def __contains__(self, key: Any) -> bool:
        return key in self._cache

    def __getitem__(self, key: Any) -> Any:
        return self._cache[key]

    def set(self, key: Any, value: Any) -> None:
        self._cache[key] = value



from typing import Protocol, TypeVar

from .job import Job

JobType = TypeVar("JobType", bound=Job)


class QueueProtocol(Protocol[JobType]):
    def put(self, job: JobType) -> None: ...
    def get_one(self) -> JobType: ...



from typing import Any, Callable, Iterator

from .workflow_step import WorkflowStep


class Workflow:
    def __init__(self, unique_name: str) -> None:
        self.unique_name = unique_name
        self._steps: list[WorkflowStep] = []

    def __getitem__(self, index: int) -> WorkflowStep:
        try:
            return self._steps[index]
        except IndexError:
            raise IndexError(
                f'Workflow "{self.unique_name}" has {len(self)} steps '
                f"- tried to access step #{index}"
            ) from None

    def __iter__(self) -> Iterator[WorkflowStep]:
        return iter(self._steps)

    def __len__(self) -> int:
        return len(self._steps)

    def step(self, callable: Callable[..., Any]) -> WorkflowStep:
        step = WorkflowStep(self, callable)
        self._steps.append(step)
        return step



from typing import Generic, TypeVar

from .exceptions import AbortJob, SkipNSteps
from .handler import ErrorHookHandler
from .interrupt import DelayedKeyboardInterrupt
from .job import Job
from .job_state_store import JobStateStoreWorkerProtocol
from .log import LOG
from .queue import QueueProtocol
from .workflow_registry import WorkflowRegistry

JobType = TypeVar("JobType", bound=Job)


class JobRunner(Generic[JobType]):
    def __init__(
        self,
        queue: QueueProtocol[JobType],
        workflow_registry: WorkflowRegistry,
        job_state_store: JobStateStoreWorkerProtocol[JobType],
        error_hook_handler: ErrorHookHandler[JobType],
    ) -> None:
        self.queue = queue
        self.workflow_registry = workflow_registry
        self.job_state_store = job_state_store
        self.error_hook_handler = error_hook_handler

    def _run_job(self, job: JobType) -> None:
        input_value = job.get_input_value()

        workflow = self.workflow_registry[job.workflow_name]
        step_to_run = workflow[job.steps_completed]

        job.mark_running(step_to_run.name)
        self.job_state_store.update(job)

        try:
            LOG.info("Running %s - input value: %s", str(step_to_run), input_value)
            with step_to_run.build_args(job.user_context, input_value) as all_args:
                args, kwargs = all_args
                retval = step_to_run(*args, **kwargs)
        except AbortJob as exc:
            LOG.info("User requested to abort job: %s", exc)
            job.mark_aborted()
        except SkipNSteps as exc:
            LOG.info("User requested to skip %d steps", exc.n)
            job.mark_n_steps_completed(exc.n + 1, exc.retval, len(workflow))
        except Exception as exc:
            LOG.exception("Job raised an exception")
            job.mark_failed(exc)
            self.error_hook_handler.notify(job, exc)
        else:
            LOG.info("Step completed successfully - return value: %s", retval)
            job.mark_n_steps_completed(1, retval, len(workflow))

        self.job_state_store.update(job)

        if job.should_be_requeued():
            LOG.info("Requeuing job")
            self.queue.put(job)

    def run(self) -> None:
        while True:
            LOG.info("Listening for next job")
            try:
                job = self.queue.get_one()
            except KeyboardInterrupt:
                return

            LOG.info("Job acquired")
            try:
                with DelayedKeyboardInterrupt():
                    self._run_job(job)
            except KeyboardInterrupt:
                return
